import { ComputedStyleStorage } from './ComputedStyleStorage.js';
import { RevealBoundaryStore } from './RevealBoundryStore.js';
declare type BorderDecoration = 'round' | 'bevel' | 'miter';
export interface CachedBoundingRect {
    top: number;
    left: number;
    width: number;
    height: number;
}
export interface CachedStyle {
    color: string;
    opacity: number;
    trueFillRadius: [number, number];
    borderStyle: string;
    borderDecorationType: BorderDecoration;
    borderDecorationRadius: number;
    topLeftBorderDecorationRadius: number;
    topRightBorderDecorationRadius: number;
    bottomLeftBorderDecorationRadius: number;
    bottomRightBorderDecorationRadius: number;
    borderWidth: number;
    fillMode: string;
    fillRadius: number;
    diffuse: boolean;
    revealAnimateSpeed: number;
    revealReleasedAccelerateRate: number;
}
interface CachedReveal {
    radius: number;
    color: string;
    opacity: number;
}
/**
 * Cached border and body pattern, for faster painting on the canvas.
 * @interface CachedRevealBitmap
 */
export interface CachedRevealBitmap {
    cachedReveal: CachedReveal;
    cachedCanvas: {
        borderReveal: HTMLCanvasElement;
        fillReveal: HTMLCanvasElement;
    };
    cachedCtx: {
        borderReveal: CanvasRenderingContext2D | null;
        fillReveal: CanvasRenderingContext2D | null;
    };
    cachedPattern: {
        borderReveal: CanvasPattern | null;
        fillReveal: CanvasPattern | null;
    };
}
export declare class CanvasConfig {
    protected _store: RevealBoundaryStore;
    pxRatio: number;
    canvas: HTMLCanvasElement;
    ctx: CanvasRenderingContext2D | null;
    computedStyle: ComputedStyleStorage;
    paintedWidth: number;
    paintedHeight: number;
    cachedImg: CachedRevealBitmap;
    currentFrameId: number;
    cachedBoundingRectFrameId: number;
    cachedBoundingRect: CachedBoundingRect;
    cachedStyleFrameId: number;
    cachedStyle: CachedStyle;
    mouseUpClientX: number | null;
    mouseUpClientY: number | null;
    mouseDownAnimateStartFrame: number | null;
    mouseDownAnimateCurrentFrame: number | null;
    mouseDownAnimateReleasedFrame: number | null;
    mouseDownAnimateLogicFrame: number | null;
    mousePressed: boolean;
    mouseReleased: boolean;
    constructor(store: RevealBoundaryStore, $el: HTMLCanvasElement);
    cacheBoundingRect: () => void;
    getTrueFillRadius: (trueFillRadius?: [number, number], fillMode?: string, fillRadius?: number) => [number, number];
    cacheCanvasPaintingStyle: () => void;
    updateCachedReveal: () => void;
    updateCachedBitmap: () => void;
    mouseInCanvas: () => boolean;
    syncSizeToElement: (x: HTMLCanvasElement) => void;
    syncSizeToRevealRadius: (x: HTMLCanvasElement) => void;
    updateAnimateGrd: (frame: number, grd: CanvasGradient) => void;
    private drawShape;
    paint: (force?: boolean | undefined, debug?: boolean | undefined) => boolean;
}
export {};
//# sourceMappingURL=CanvasConfig.d.ts.map